import pprint
from typing import Optional

from langevals_core.base_evaluator import (
    BaseEvaluator,
    EvaluationResult,
    EvaluatorEntry,
)
from pydantic import BaseModel

from langevals.utils import get_evaluator_definitions


class Expectation(BaseModel):
    entry: dict[str, str] = {}

    def _build_entry(self, evaluator: BaseEvaluator):
        evaluator_definitions = get_evaluator_definitions(evaluator)
        return evaluator_definitions.entry_type(**self.entry)

    def evaluate(self, evaluator: BaseEvaluator):
        entry = self._build_entry(evaluator)

        return evaluator.evaluate(entry)

    def to_pass(self, evaluator: BaseEvaluator):
        entry = self._build_entry(evaluator)
        result = self.evaluate(evaluator)
        assert result.status == "processed", (
            result.details
            if result.status == "skipped"
            else result.message if result.status == "error" else None
        )
        assert result.passed, self._result_details_print(
            entry, f"{evaluator.name} to_pass FAILED", result.details
        )

    def to_fail(self, evaluator: BaseEvaluator):
        entry = self._build_entry(evaluator)
        result = self.evaluate(evaluator)
        assert result.status == "processed", (
            result.details
            if result.status == "skipped"
            else result.message if result.status == "error" else None
        )
        assert not result.passed, self._result_details_print(
            entry, f"{evaluator.name} to_fail FAILED", result.details
        )

    def _result_details_print(
        self, entry: EvaluatorEntry, msg: str, details: Optional[str]
    ):
        return (
            " - ".join([x for x in [msg, details] if x])
            + "\nEntry: "
            + pprint.pformat(entry)
        )

    def score(self, evaluator: BaseEvaluator):
        entry = self._build_entry(evaluator)
        result = self.evaluate(evaluator)
        assert result.status == "processed", (
            result.details
            if result.status == "skipped"
            else result.message if result.status == "error" else None
        )
        return NumericExpectation(entry=entry, evaluator=evaluator, result=result)


class NumericExpectation(BaseModel):
    entry: EvaluatorEntry
    evaluator: BaseEvaluator
    result: EvaluationResult

    def _result_details_print(self, msg, value):
        return (
            " - ".join(
                [
                    x
                    for x in [
                        f"{self.evaluator.name} {msg} {value} FAILED (actual: {self.result.score})",
                        self.result.details,
                    ]
                    if x
                ]
            )
            + "\nEntry: "
            + pprint.pformat(self.entry)
        )

    def to_be_greater_than(self, value: float):
        assert self.result.score > value, self._result_details_print(
            "to_be_greater_than", value
        )

    def to_be_less_than(self, value: float):
        assert self.result.score < value, self._result_details_print(
            "to_be_less_than", value
        )

    def to_be_equal_to(self, value: float):
        assert self.result.score == value, self._result_details_print(
            "to_be_equal_to", value
        )

    def to_be_within(self, lower: float, upper: float):
        assert lower <= self.result.score <= upper, self._result_details_print(
            "to_be_within", (lower, upper)
        )

    def to_be_greater_than_or_equal(self, value: float):
        assert self.result.score >= value, self._result_details_print(
            "to_be_greater_or_equal", value
        )

    def to_be_less_than_or_equal(self, value: float):
        assert self.result.score <= value, self._result_details_print(
            "to_be_less_than_or_equal"
        )


def expect(
    input: Optional[str] = None,
    output: Optional[str] = None,
    contexts: Optional[list[str]] = None,
    expected_output: Optional[str] = None,
):
    entry = {}
    if input is not None:
        entry["input"] = input
    if output is not None:
        entry["output"] = output
    if contexts is not None:
        entry["contexts"] = contexts
    if expected_output is not None:
        entry["expected_output"] = expected_output

    return Expectation(entry=entry)



import importlib
import importlib.metadata
import pkgutil
import textwrap
from typing import Optional, Type, get_args

from langevals_core.base_evaluator import (
    BaseEvaluator,
    EvaluationResult,
    EvaluatorEntry,
)
from pydantic import BaseModel


def load_evaluator_packages():
    evaluators = {}
    for distribution in importlib.metadata.distributions():
        normalized_name = distribution.metadata["Name"].replace("-", "_")
        if normalized_name == "langevals_core":
            continue
        if normalized_name.startswith("langevals_"):
            try:
                evaluators[normalized_name] = importlib.import_module(normalized_name)
            except ImportError:
                pass
    return evaluators


def get_evaluator_classes(evaluator_package):
    evaluator_classes: list[BaseEvaluator] = []
    package_path = evaluator_package.__path__
    for _, module_name, _ in pkgutil.walk_packages(package_path):
        print(f"Loading {evaluator_package.__name__}.{module_name}")
        module = importlib.import_module(f"{evaluator_package.__name__}.{module_name}")
        for name, cls in module.__dict__.items():
            if (
                isinstance(cls, type)
                and issubclass(cls, BaseEvaluator)
                and cls is not BaseEvaluator
            ):
                evaluator_classes.append(cls)  # type: ignore

    return evaluator_classes


class EvaluatorDefinitions(BaseModel):
    module_name: str
    evaluator_name: str
    entry_type: Type[EvaluatorEntry]
    settings_type: Type[BaseModel]
    result_type: Type[EvaluationResult]
    name: str
    description: str
    docs_url: Optional[str]
    env_vars: list[str]
    is_guardrail: bool
    category: str


def get_evaluator_definitions(evaluator_cls: BaseEvaluator):
    fields = evaluator_cls.model_fields

    settings_type = fields["settings"].annotation
    entry_type = get_args(fields["entry"].annotation)[0]
    result_type = get_args(fields["result"].annotation)[0]

    module_name, evaluator_name = evaluator_cls.__module__.split(".", 1)
    module_name = module_name.split("langevals_")[1]

    if getattr(evaluator_cls, "name", None) is None:
        raise ValueError(f"Missing name attribute in {evaluator_cls}")

    name = evaluator_cls.name
    docs_url = evaluator_cls.docs_url
    description = textwrap.dedent(evaluator_cls.__doc__ or "")
    docs_url = evaluator_cls.docs_url
    env_vars = evaluator_cls.env_vars
    is_guardrail = evaluator_cls.is_guardrail
    category = evaluator_cls.category

    return EvaluatorDefinitions(
        module_name=module_name,
        evaluator_name=evaluator_name,
        entry_type=entry_type,
        settings_type=settings_type,  # type: ignore
        result_type=result_type,
        name=name,
        description=description,
        docs_url=docs_url,
        env_vars=env_vars,
        is_guardrail=is_guardrail,
        category=category,
    )



import pytest


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "pass_rate: allows for passing the tests even if they partially fail less than the rate, with 1.0 meaning all samples must pass, and 0.0 meaning all samples are allowed to fail.",
    )


def pytest_collection_modifyitems(session, config, items):
    session.pass_rates = {}
    for item in items:
        marker = item.get_closest_marker("pass_rate")
        if marker:
            group_name = item.nodeid.split("[")[0]
            if group_name not in session.pass_rates:
                session.pass_rates[group_name] = {
                    "total": 0,
                    "passed": set(),
                    "required_rate": marker.args[0],
                    "processed": set(),
                    "reports": {},
                }
            session.pass_rates[group_name]["total"] += 1


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield

    report = outcome.get_result()

    if report.when == "call":  # Only consider the call phase
        group_name = item.nodeid.split("[")[0]
        if group_name in item.session.pass_rates:
            if report.outcome == "failed":
                pass_rate_data = item.session.pass_rates[group_name]
                optimistic_pass_rate = (
                    pass_rate_data["total"]
                    - len(pass_rate_data["processed"])
                    + len(pass_rate_data["passed"])
                )
                current_rate = optimistic_pass_rate / pass_rate_data["total"]
                required_rate = pass_rate_data["required_rate"]

                # Check if the group pass rate is still acceptable
                if current_rate >= required_rate:
                    # We manipulate the report here to change its outcome
                    report.outcome = "skipped"
                    report.longrepr = (
                        f"acceptable failure (under {required_rate:.2%} pass rate)"
                    )
                    report.wasxfail = "reason: under acceptable failure rate"
            # If a retry mechanism gets the test to eventually pass, we mark it all as passed
            elif (
                report.outcome == "passed"
                and item.nodeid in item.session.pass_rates[group_name]["reports"]
            ):
                previous_reports = item.session.pass_rates[group_name]["reports"][
                    item.nodeid
                ]
                for previous_report in previous_reports:
                    if previous_report.outcome == "failed":
                        report.outcome = "passed"
                        report.longrepr = f"passed eventually"

            if item.nodeid not in item.session.pass_rates[group_name]["reports"]:
                item.session.pass_rates[group_name]["reports"][item.nodeid] = []
            item.session.pass_rates[group_name]["reports"][item.nodeid].append(report)

    report.session = item.session


def pytest_runtest_logreport(report):
    if report.when == "call":
        group_name = report.nodeid.split("[")[0]
        if group_name in report.session.pass_rates:

            report.session.pass_rates[group_name]["processed"].add(report.nodeid)
            if report.outcome == "passed":
                report.session.pass_rates[group_name]["passed"].add(report.nodeid)
            elif report.nodeid in report.session.pass_rates[group_name]["passed"]:
                report.session.pass_rates[group_name]["passed"].remove(report.nodeid)


def pytest_sessionfinish(session, exitstatus):
    for group_name, data in session.pass_rates.items():
        pass_rate = len(data["passed"]) / data["total"]
        required_rate = data["required_rate"]
        if pass_rate < required_rate:
            print(
                f"\n\nTest group '{group_name}' failed to meet the pass rate requirement: {pass_rate:.2%} / {required_rate:.2%} required."
            )
            session.exitstatus = 1  # Set exit status to failure if any group fails



import sys
import dotenv

from langevals.utils import (
    get_evaluator_classes,
    get_evaluator_definitions,
    load_evaluator_packages,
)

dotenv.load_dotenv()

import asyncio
from fastapi import FastAPI, HTTPException, Request
from typing import List, Optional
from langevals_core.base_evaluator import (
    EvaluationResultSkipped,
    EvaluationResultError,
)
from pydantic import BaseModel, ConfigDict, Field, ValidationError
from mangum import Mangum

app = FastAPI()


def create_evaluator_routes(evaluator_cls):
    definitions = get_evaluator_definitions(evaluator_cls)
    module_name = definitions.module_name
    evaluator_name = definitions.evaluator_name
    entry_type = definitions.entry_type
    settings_type = definitions.settings_type
    result_type = definitions.result_type

    required_env_vars = (
        "\n\n__Env vars:__ " + ", ".join(definitions.env_vars)
        if len(definitions.env_vars) > 0
        else ""
    )
    docs_url = "\n\n__Docs:__ " + definitions.docs_url if definitions.docs_url else ""
    description = definitions.description + required_env_vars + docs_url

    class Request(BaseModel):
        model_config = ConfigDict(extra="forbid")

        data: List[entry_type] = Field(description="List of entries to be evaluated, check the field type for the necessary keys")  # type: ignore
        settings: Optional[settings_type] = Field(None, description="Evaluator settings, check the field type for what settings this evaluator supports")  # type: ignore
        env: Optional[dict[str, str]] = Field(
            None,
            description="Optional environment variables to override the server ones",
            json_schema_extra={"example": {}},
        )

    evaluator_cls.preload()

    @app.post(
        f"/{module_name}/{evaluator_name}/evaluate",
        name=f"{module_name}_{evaluator_name}_evaluate",
        description=description,
    )
    async def evaluate(
        req: Request,
    ) -> List[result_type | EvaluationResultSkipped | EvaluationResultError]:  # type: ignore
        evaluator = evaluator_cls(settings=(req.settings or {}), env=req.env)  # type: ignore
        return evaluator.evaluate_batch(req.data)


evaluators = load_evaluator_packages()
for evaluator_name, evaluator_package in evaluators.items():
    module_name = evaluator_package.__name__.split("langevals_")[1]
    if (
        len(sys.argv) > 2
        and sys.argv[1] == "--only"
        and module_name not in sys.argv[2].split(",")
    ):
        continue
    print(f"Loading {evaluator_package.__name__}")
    for evaluator_cls in get_evaluator_classes(evaluator_package):
        create_evaluator_routes(evaluator_cls)


@app.exception_handler(ValidationError)
async def validation_exception_handler(request: Request, exc: ValidationError):
    raise HTTPException(
        status_code=400,
        detail=exc.errors(),
    )


def main():
    if len(sys.argv) > 1 and sys.argv[1] == "--preload":
        print("Preloading done")
        return
    if len(sys.argv) > 1 and sys.argv[1] == "--export-openapi-json":
        import json

        with open("openapi.json", "w") as f:
            f.write(json.dumps(app.openapi(), indent=2))
        print("openapi.json exported")
        return
    from hypercorn.config import Config
    from hypercorn.asyncio import serve

    asyncio.run(serve(app, Config()))  # type: ignore


if __name__ == "__main__":
    main()
else:
    handler = Mangum(app, lifespan="off")



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional
from langevals_core.base_evaluator import (
    BaseEvaluator,
    BatchEvaluationResult,
    EvaluationResult,
    EvaluatorEntry,
)
import pandas as pd
from pydantic import BaseModel

from langevals.utils import get_evaluator_definitions


class GenericEvaluatorEntry(EvaluatorEntry):
    input: Optional[str] = None
    output: Optional[str] = None
    contexts: Optional[list[str]] = None
    expected_output: Optional[str] = None


class EvaluationResultSet(BaseModel):
    entries: list[EvaluatorEntry]
    evaluators: list[BaseEvaluator]
    results: list[BatchEvaluationResult]

    def to_list(self):
        result = {}
        for i, evaluator in enumerate(self.evaluators):
            evaluator_definitions = get_evaluator_definitions(evaluator)
            result[evaluator_definitions.evaluator_name] = [
                result.model_dump() for result in self.results[i]
            ]

        return result

    def to_pandas(self):
        records = {}
        for entry in self.entries:
            for key in entry.model_dump().keys():
                records[key] = []

        entry_keys = records.keys()
        for i, entry in enumerate(self.entries):
            for key in entry_keys:
                records[key].append(entry.model_dump().get(key, None))

        for i, evaluator in enumerate(self.evaluators):
            evaluator_definitions = get_evaluator_definitions(evaluator)
            for result in self.results[i]:
                result_dict = result.model_dump()
                passed = result_dict.get("passed", None)
                score = result_dict.get("score", None)
                status = result_dict.get("status", None)
                details = result_dict.get("details", result_dict.get("message", None))

                if evaluator_definitions.evaluator_name not in records:
                    records[evaluator_definitions.evaluator_name] = []
                records[evaluator_definitions.evaluator_name].append(
                    status
                    if status != "processed"
                    else passed if passed is not None else score
                )

                details_column = f"{evaluator_definitions.evaluator_name}_details"
                if details is not None and details_column not in records:
                    records[details_column] = [None] * (
                        len(records[evaluator_definitions.evaluator_name]) - 1
                    )
                if details_column in records:
                    records[details_column].append(details)

        df = pd.DataFrame(records)
        df = df.dropna(axis=1, how="all")

        return df


# TODO: docs, and auto-generated docs from evaluators
def evaluate(
    entries: pd.DataFrame, evaluators: list[BaseEvaluator]
) -> EvaluationResultSet:
    entries_ = _pandas_to_generic_entries(entries)
    result_set: list[BatchEvaluationResult] = [[]] * len(evaluators)

    with ThreadPoolExecutor(max_workers=5) as executor:
        future_to_index = {
            executor.submit(evaluator.evaluate_batch, entries_): idx
            for idx, evaluator in enumerate(evaluators)
        }

        for future in as_completed(future_to_index):
            idx = future_to_index[future]
            result_set[idx] = future.result()

    return EvaluationResultSet(
        entries=entries_, evaluators=evaluators, results=result_set
    )


def _pandas_to_generic_entries(entries: pd.DataFrame) -> list[EvaluatorEntry]:
    return [GenericEvaluatorEntry(**entry.to_dict()) for _, entry in entries.iterrows()]
